/*
** ATOP - System & Process Monitor 
**
** The program 'atop' offers the possibility to view the activity of
** the system on system-level as well as process-level.
** 
** This source-file contains the print-functions to visualize the calculated
** figures.
** ================================================================
** Author:      Gerlof Langeveld - AT Computing, Nijmegen, Holland
** E-mail:      gerlof@ATComputing.nl
** Date:        November 1996
** LINUX-port:  June 2000
** 
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License as published by the
** Free Software Foundation; either version 2, or (at your option) any
** later version.
** 
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
** See the GNU General Public License for more details.
**
** $Log: showgeneric.c,v $
** Revision 1.9  2001/11/29 10:41:44  gerlof
** *** empty log message ***
**
** Revision 1.8  2001/11/29 10:38:16  gerlof
** Exit-code correctly printed.
**
** Revision 1.7  2001/11/26 11:18:45  gerlof
** Modified generic output in case that the kernel-patch is not installed.
**
** Revision 1.6  2001/11/13 08:24:50  gerlof
** Show blank columns for sockets and disk I/O when no kernel-patch installed.
**
** Revision 1.5  2001/11/07 09:19:28  gerlof
** Use /proc instead of /dev/kmem for process-level statistics.
**
** Revision 1.4  2001/10/05 13:46:32  gerlof
** Implemented paging through the process-list
**
** Revision 1.3  2001/10/04 08:47:27  gerlof
** Improved handling of error-messages
**
** Revision 1.2  2001/10/03 08:57:53  gerlof
** Improved help-screen shown in scrollable window
**
** Revision 1.1  2001/10/02 10:43:34  gerlof
** Initial revision
**
*/

static char rcsid[] = "$Id: showgeneric.c,v 1.9 2001/11/29 10:41:44 gerlof Exp $";

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <signal.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <termio.h>
#include <unistd.h>
#include <stdarg.h>
#include <curses.h>
#include <pwd.h>
#include <grp.h>

#include "atop.h"

#define	MINLINES	12
#define	ALLUSERS	9999999

extern ulong	interval;
extern int	deviatonly;
extern char	kernpatch;
extern int	acctfd;

extern char	flaglist[];

struct syscap {
	ulong	availcpu;
	ulong	availmem;
	ulong	availdsk;
	ulong	availnet;
};

/*
** format-lines for printing
*/
char	*genhdr  	= "ATOP - System & Process Monitor   "
		     	  "      %s   %12ld seconds elapsed\n";

char 	*prcline   	= "PRC | sys %s | user %s | #procs %s | #exits %s |"
	       	     	  " #zombie %s |\n";


char	*genprochdr1	= "\n  PID  SYSTEM    USER  VGROW  RGROW  USERNAME  PRI"
		     	  " EXCODE S  %s CMD    %3d/%3d\n";
char	*genprochdr2	= "\n  PID  SYSTEM    USER  VGROW  RGROW  READ WRITE "
		     	  "RNET SNET S  %s CMD    %3d/%3d\n";

char 	*genproclina1	= "%5d %s %s %s %s  %-9.9s %3d        %c %3.0f%% %.14s\n";
char 	*genproclina2	= "%5d %s %s %s %s %s %s %s %s %c %3.0f%% %.14s\n";

char 	*genprocline1	= "%s %s                %-9.9s     %s"
			  " E %3.0f%% <%.12s>\n";
char 	*genprocline2	= "%s %s               (r&w %s)        "
			  "  E %3.0f%% <%.12s>\n";


char	*memprochdr	= "\n  PID MINFLT MAJFLT #SWP VSTEXT "
		     	  " VSIZE  RSIZE  VGROW  RGROW  %s CMD    %3d/%3d\n";
char 	*memproclina	= "%5d %s %s %s %s %s %s %s %s %3.0f%%";
char 	*memprocline	= "%s %s %s                                  "
			  "  %3.0f%% <%.12s>\n";


char	*netprochdr	= "\n  PID   TCPRCV  TCPSND    UDPRCV  UDPSND"
		     	  "                     %s CMD    %3d/%3d\n";
char 	*netproclina	= "%5d  %s %s   %s %s                    %3.0f%%";
char 	*netprocline	= "                                        "
			  "              %3.0f%% <%.12s>\n";


char	*staprochdr	= "\n  PID  USERNAME GROUP    EXCODE   "
		     	  " STDATE    STTIME       S  %s CMD    %3d/%3d\n";
char 	*staproclina	= "%5d  %-8s %-8s         %s %s      %c %3.0f%%";
char 	*staprocline	= " %-8s %-8s %s  %s %s      E %3.0f%% <%.12s>\n";

uid_t	uidselected 	= ALLUSERS; /* selected user to be shown        */
char	namselected[256];

char	procnamesz;
char	procname[64];

static int	screen;     	/* boolean: screen control-codes needed */
static int	paused;     	/* boolean: currently in pause-mode     */
static int	fixedhead;	/* boolean: fixate header-lines         */


/*
** internal proto-types
*/
static int	prisyst(struct sstat *,  int, int);
static int	priproc(struct pstat  *, int, int, int, int, int,
		        char *, struct syscap *, char);

static void	showgenproc(struct pstat *, double);
static void	showmemproc(struct pstat *, double);
static void	shownetproc(struct pstat *, double);
static void	showstaproc(struct pstat *, double);

static int	printg(const char *, ...);
static void	showhelp(int);


/*
** print the deviation-counters on process- and system-level
*/
void
generic_samp(time_t curtime, int nsecs,
           struct sstat *sstat, struct pstat *pstat,
           int nact, int nproc, int nzomb, int nexit)
{
	static int	callnr = 0;

	static int	(*curcompar)(const void *, const void *) = compcpu;
	static char	*sortorder = "CPU", showtype = 'g';

	struct syscap	syscap;
	register int	i, curline, statline;
	register ulong	totut, totst;
	int		firstproc, plistsz;
	char		format1[16], format2[16], format3[16];
	char		format4[16], format5[16];

	struct passwd 	*pwd;

	if (callnr == 0)	/* first call ? */
	{
		/*
		** check if default sortorder and/or showtype are overruled
		** by commando-line flags
		*/
		for (i=0; flaglist[i]; i++)
		{
			switch (flaglist[i])
			{
			   case 'C':
				curcompar = compcpu;
				sortorder = "CPU";
				break;
			   case 'M':
				curcompar = compmem;
				sortorder = "MEM";
				break;
			   case 'D':
				curcompar = compdsk;
				sortorder = "DSK";
				break;
			   case 'N':
				curcompar = compnet;
				sortorder = "NET";
				break;
			   case 'g':
				showtype  = 'g';
				curcompar = compcpu;
				sortorder = "CPU";
				break;
			   case 'm':
				showtype  = 'm';
				curcompar = compmem;
				sortorder = "MEM";
				break;
			   case 'n':
				if (!kernpatch)
				{
					fprintf(stderr,
						"No kernel-patch installed "
						"(no network-statistics)\n");
					sleep(3);
					break;
				}

				showtype  = 'n';
				curcompar = compnet;
				sortorder = "NET";
				break;
			   case 's':
				showtype  = 's';
				break;

			   case 'a':
				if (deviatonly)
					deviatonly=0;
				else
					deviatonly=1;
				break;

			   case 'f':
				if (fixedhead)
					fixedhead=0;
				else
					fixedhead=1;
				break;

			   default:
				prusage("atop");
			}
		}

        	/*
        	** set stdout output on line-basis
        	*/
        	setvbuf(stdout, (char *)0, _IOLBF, BUFSIZ);

        	/*
        	** check if STDOUT is related to a tty or
        	** something else (file, pipe)
        	*/
        	if ( isatty(1) )
                	screen = 1;
        	else
                	screen = 0;

        	/*
        	** install catch-routine to finish in a controlled way
		** and activate cbreak-mode
        	*/
        	if (screen)
		{
			/*
			** initialize screen-handling via curses
			*/
			initscr();
			cbreak();
			noecho();

			if (LINES < MINLINES)
			{
				printw("Not enough screen-lines "
				       "(need at least %d lines)\n", MINLINES);
				refresh();
				sleep(3);
				cleanstop(1);
			}

			if (COLS  < 80)
			{
				printw("Not enough columns "
				       "(need at least %d columns)\n", 80);
				refresh();
				sleep(3);
				cleanstop(1);
			}
		} 
                signal(SIGINT,   cleanstop);
                signal(SIGTERM,  cleanstop);
	}

	callnr++;

	/*
	** calculate the total consumption on system-level for the 
	** four main resources
	*/
	syscap.availcpu	= sstat->cpu.stime + sstat->cpu.utime +
			  sstat->cpu.ntime + sstat->cpu.itime;
	syscap.availmem	= sstat->mem.physmem * NBPP/1024;
	syscap.availnet	= sstat->net.tcpin + sstat->net.tcpout +
	                  sstat->net.udpin + sstat->net.udpout;

	/*
	** calculate total number of accesses which have been
	** issued by the active disks
	*/
	syscap.availdsk = 0;

	for (i=0; sstat->xdsk[i].name[0]; i++)
		syscap.availdsk += sstat->xdsk[i].nread+sstat->xdsk[i].nwrite;

	if (syscap.availcpu == 0)	syscap.availcpu = 1;
	if (syscap.availmem == 0)	syscap.availmem = 1;
	if (syscap.availnet == 0)	syscap.availnet = 1;
	if (syscap.availdsk == 0)	syscap.availdsk = 1;

        /*
        ** prepare screen or file output for new sample
        */
        if (screen)
                clear();
        else
                printf("\n\n");

        /*
        ** print general headerlines
        */
        convtime(curtime, format1);       /* convert time to ascii string   */

	if (screen)
		attron(A_REVERSE);

       	printg(genhdr, format1, nsecs);

	curline = 1;

	if (screen)
		attroff(A_REVERSE);

	/*
	** calculate cumulative system- and user-time for all active processes
	*/
	for (i=totut=totst=0; i < nact; i++)
	{
		totut += (pstat+i)->cpu.utime;
		totst += (pstat+i)->cpu.stime;
	}

	printg(prcline, val2cpustr(totst * 1000/HZ, format1),
			val2cpustr(totut * 1000/HZ, format2),
			val2valstr(nproc,           format3, 5),
			acctfd >= 0 ?
			          val2valstr(nexit, format4, 5) : "    ?",
			val2valstr(nzomb,           format5, 4));

	curline++;

	curline  = prisyst(sstat, curline, nsecs);
	statline = curline;

	if (screen)
		plistsz = LINES-curline-2;
	else
		plistsz = nact;

	/*
	** loop in which the list of active processes is shown;
	** the loop will be preempted by receiving a timer-signal
 	** or when the trigger-button is pressed.
	*/
	firstproc = 0;

	while (1)
	{
		/*
		** sort the list of active processes in required order 
		** (default CPU-consumption) and print the list
		*/
		if (nact > 0)
		{
			qsort  (pstat, nact, sizeof(struct pstat), curcompar);
			priproc(pstat, firstproc, nact, curline,
			        firstproc/plistsz+1, (nact-1)/plistsz+1,
				         sortorder, &syscap, showtype);
		}

		/*
		** in case of writing to a terminal, the user can also enter
		** a character to switch options, etc
		*/
		if (screen)
		{
			int  n;

			/*
			** await input-character or interval-timer expiration
			*/
			if (paused)
			{
				move(statline, 73);
				attron(A_BLINK);
				attron(A_REVERSE);
				printw("PAUSED");
				attroff(A_REVERSE);
				attroff(A_BLINK);
			}

			switch ( mvgetch(statline, 0) )
			{
			   /*
			   ** timer expired
			   */
			   case ERR:
			   case 0:
				return;	

			   /*
			   ** stop it
			   */
			   case 'q':
				cleanstop(0);

			   /*
			   ** manual trigger for next sample
			   */
			   case 't':
				if (paused)
					break;

				getusr1();
				return;

			   /*
			   ** sort in cpu-activity order
			   */
			   case 'C':
				curcompar = compcpu;
				sortorder = "CPU";
				firstproc = 0;
				break;

			   /*
			   ** sort in memory-consumption order
			   */
			   case 'M':
				curcompar = compmem;
				sortorder = "MEM";
				firstproc = 0;
				break;

			   /*
			   ** sort in disk-activity order
			   */
			   case 'D':
				curcompar = compdsk;
				sortorder = "DSK";
				firstproc = 0;
				break;

			   /*
			   ** sort in network-activity order
			   */
			   case 'N':
				curcompar = compnet;
				sortorder = "NET";
				firstproc = 0;
				break;

			   /*
			   ** general figures per process
			   */
			   case 'g':
				showtype  = 'g';
				curcompar = compcpu;
				sortorder = "CPU";
				firstproc = 0;
				break;

			   /*
			   ** memory-specific figures per process
			   */
			   case 'm':
				showtype  = 'm';
				curcompar = compmem;
				sortorder = "MEM";
				firstproc = 0;
				break;

			   /*
			   ** network-specific figures per process
			   */
			   case 'n':
				if (!kernpatch)
				{
					printw("No kernel-patch installed; "
					       "request ignored!");
					break;
				}
				showtype  = 'n';
				curcompar = compnet;
				sortorder = "NET";
				firstproc = 0;
				break;

			   /*
			   ** static info per process
			   */
			   case 's':
				showtype  = 's';
				firstproc = 0;
				break;

			   /*
			   ** help wanted?
			   */
			   case '?':
			   case 'h':
			   case 'H':
				alarm(0);	/* stop the clock         */

				move(statline, 0);
				clrtobot();	/* blank the process-list */
				refresh();

				showhelp(statline+1);

				move(statline, 0);

				if (interval && !paused)
					alarm(1); /* force new sample     */

				firstproc = 0;
				break;

			   /*
			   ** change interval timeout
			   */
			   case 'i':
				alarm(0);	/* stop the clock */
				echo();

				move(statline, 0);
				clrtoeol();
				printw("New interval (in seconds): ");

				n = -1;
				scanw("%d", &n);

				noecho();

				if (n < 0)
				{
					beep();
					move(statline, 0);
					clrtoeol();
					printw("Illegal value!");
					refresh();
					sleep(1);
				}
				else
					interval = n;

				move(statline, 0);

				if (interval)
				{
					if (!paused)
						alarm(1); /* set short timer */
				}
				else
				{
					clrtoeol();
					printw("No timer set; waiting for "
					       "manual trigger ('t').....");
					refresh();
				}

				firstproc = 0;
				break;

			   /*
			   ** focus on specific user
			   */
			   case 'u':
				alarm(0);	/* stop the clock */
				echo();

				move(statline, 0);
				clrtoeol();
				printw("Specify user-name (enter=all users): ");

				namselected[0] = '\0';
				scanw("%255s\n", namselected);

				noecho();

				if (namselected[0])	/* data entered ? */
				{
					if (pwd = getpwnam(namselected))
					{
						uidselected = pwd->pw_uid;
					}
					else
					{
						uidselected = ALLUSERS;

						beep();
						move(statline, 0);
						clrtoeol();
						printw("User unknown!");
						refresh();
						sleep(1);
					}
				}
				else
					uidselected = ALLUSERS;

				move(statline, 0);

				if (interval && !paused)
					alarm(1);  /* set short timer */

				firstproc = 0;
				break;

			   /*
			   ** focus on specific process-name
			   */
			   case 'P':
				alarm(0);	/* stop the clock */
				echo();

				move(statline, 0);
				clrtoeol();
				printw("Specify (prefix of) process-name "
				       "(enter=all processes): ");

				procname[0] = '\0';
				scanw("%63s\n", procname);

				procnamesz = strlen(procname);

				noecho();

				move(statline, 0);

				if (interval && !paused)
					alarm(1);  /* set short timer */

				firstproc = 0;
				break;

			   /*
			   ** toggle pause-state
			   */
			   case 'p':
				if (paused)
				{
					paused=0;
					clrtoeol();
					refresh();
					alarm(1);	 /* set short timer */
				}
				else
				{
					paused=1;
					clrtoeol();
					refresh();
					alarm(0);	/* stop the clock */
				}
				break;

			   /*
			   ** toggle between modified processes and
			   ** all processes
			   */
			   case 'a':
				if (deviatonly)
				{
					deviatonly=0;
					printw("All processes will be shown "
					       "......");
				}
				else
				{
					deviatonly=1;
					printw("Only active processes will"
					       " be shown.....");
				}

				if (interval && !paused)
					alarm(1);  /* set short timer */

				firstproc = 0;
				break;

			   /*
			   ** system-statistics lines:
			   **	         toggle fixed or variable
			   */
			   case 'f':
				if (fixedhead)
				{
					fixedhead=0;
					printw("Only active system-resources "
					       "will be shown ......");
				}
				else
				{
					fixedhead=1;
					printw("All lines related to system-"
					       "resources will be shown.....");
				}

				firstproc = 0;
				break;

			   /*
			   ** handle forward
			   */
			   case 0x06:
				if (nact-firstproc > plistsz)
					firstproc += plistsz;
				break;

			   /*
			   ** handle backward
			   */
			   case 0x02:
				if (firstproc >= plistsz)
					firstproc -= plistsz;
				break;

			   /*
			   ** unknown key-stroke
			   */
			   default:
			        beep();
			}
		}
		else	/* no screen */
			return;
	}
}

/*
** print the list of processes from the deviation-list
*/
static int
priproc(struct pstat  *pstat, int firstproc, int lastproc,
                              int curline,   int curlist, int totlist,
        char *sortorder, struct syscap *sb, char showtype)
{
	register int		i, j;
	register struct pstat	*curstat;
	double			perc;

	/*
	** print process-list header line
	*/
	if (screen)
		attron(A_REVERSE);

	switch (showtype)
	{
	   case 'g':
		if (!kernpatch)
			printg(genprochdr1, sortorder, curlist, totlist);
		else
			printg(genprochdr2, sortorder, curlist, totlist);
		break;

	   case 'm':
		printg(memprochdr, sortorder, curlist, totlist);
		break;

	   case 'n':
		printg(netprochdr, sortorder, curlist, totlist);
		break;

	   case 's':
		printg(staprochdr, sortorder, curlist, totlist);
		break;
	}

	if (screen)
	{
		attroff(A_REVERSE);
		clrtobot();
	}

	curline += 2;

	/*
	** print info per actual process and maintain totals
	*/
	for (i=firstproc; i < lastproc; i++)
	{
		if (screen && curline >= LINES)	/* screen filled entirely ? */
			break;

		curstat = pstat+i;

		/*
		** calculate occupation-percentage of this process
		** depending on selected resource
		*/
		switch (*sortorder) 
		{
		   case 'C':
			perc = (double)(curstat->cpu.stime+curstat->cpu.utime)
			        			 * 100.0 / sb->availcpu;
			break;

		   case 'M':
			perc = (double)(curstat->mem.rmem)
                                                         * 100.0 / sb->availmem;
			break;

		   case 'D':
			perc = (double)(curstat->dsk.ior + curstat->dsk.iow)
			        			 * 100.0 / sb->availdsk;
			break;

		   case 'N':
			perc = (double)(curstat->net.tcpsnd +
			                curstat->net.tcprcv +
			                curstat->net.udpsnd +
			                curstat->net.udprcv  )
			        		         * 100.0 / sb->availnet;
			break;

		   default:
			perc = 0.0;
		}

		if (perc > 100.0)
			perc = 100.0;

		/*
		** check if only the processes of a particular user
		** should be shown
		*/
		if (uidselected != ALLUSERS && uidselected != curstat->gen.uid)
			continue;

		/*
		** check if only the processes with a particular name
		** should be shown
		*/
		if (procnamesz &&
		    memcmp(procname, curstat->gen.name, procnamesz) )
			continue;

		/*
		** now do the formatting of output
		*/
		switch (showtype)
		{
		   case 'g':
			showgenproc(curstat, perc);
			break;

		   case 'm':
			showmemproc(curstat, perc);
			break;

		   case 'n':
			shownetproc(curstat, perc);
			break;

		   case 's':
			showstaproc(curstat, perc);
			break;
		}

		curline++;
	}

	return curline;
}

/*
** print a line of general figures about one process
*/
static void
showgenproc(struct pstat *curstat, double perc)
{
	struct passwd 	*pwd;
	char		*username;

	char		format1[16], format2[16];
	char		format3[16], format4[16];
	char		format5[16], format6[16];
	char		format7[16], format8[16];

	if (curstat->gen.state != 'E')
	{
		/*
		** show process-info of active process
		*/
		if (!kernpatch)
		{
			/*
			** determine the user-name of process
			*/
			if ( (pwd = getpwuid(curstat->gen.uid)) == NULL)
				username = "unknown";
			else
				username = pwd->pw_name;

			printg(genproclina1,
			   curstat->gen.pid,
			   val2cpustr(curstat->cpu.stime*1000/HZ, format1),
			   val2cpustr(curstat->cpu.utime*1000/HZ, format2),
			   val2memstr(curstat->mem.vgrow*1024,    format3,
								  KBFORMAT),
			   val2memstr(curstat->mem.rgrow*1024,    format4,
								  KBFORMAT),
			   username,
			   curstat->cpu.pri,
			   curstat->gen.state, perc, curstat->gen.name);
		}
		else
		{
			printg(genproclina2,
			   curstat->gen.pid,
			   val2cpustr(curstat->cpu.stime*1000/HZ, format1),
			   val2cpustr(curstat->cpu.utime*1000/HZ, format2),
			   val2memstr(curstat->mem.vgrow*1024,    format3,
								  KBFORMAT),
			   val2memstr(curstat->mem.rgrow*1024,    format4,
								  KBFORMAT),
			   val2valstr(curstat->dsk.ior,           format5, 5),
			   val2valstr(curstat->dsk.iow,           format6, 5),
			   val2valstr(curstat->net.tcprcv+curstat->net.udprcv,
			                                          format7, 4),
			   val2valstr(curstat->net.tcpsnd+curstat->net.udpsnd,
			                                          format8, 4),
			   curstat->gen.state, perc, curstat->gen.name);
		}
	}
	else
	{
		/*
		** show process-info of exited process
		*/
		if (curstat->gen.pid == 0)
			printg("    ? ");
		else
			printg("%5d ", curstat->gen.pid);

		if (!kernpatch)
		{
			if ( (pwd = getpwuid(curstat->gen.uid)) == NULL)
				username = "unknown";
			else
				username = pwd->pw_name;

		  	if ( curstat->gen.excode&0xff )
				sprintf(format3, "sig_%02d",
					curstat->gen.excode&0x7f);
			else
			  	val2valstr(curstat->gen.excode>>8, format3, 6);

			printg(genprocline1,
			  val2cpustr(curstat->cpu.stime*1000/HZ, format1),
			  val2cpustr(curstat->cpu.utime*1000/HZ, format2),
			  username,
			  format3,
			  perc, curstat->gen.name);
		}
		else
		{
			printg(genprocline2,
			  val2cpustr(curstat->cpu.stime*1000/HZ, format1),
			  val2cpustr(curstat->cpu.utime*1000/HZ, format2),
			  val2valstr(curstat->dsk.ior, format3, 6),
			  perc, curstat->gen.name);
		}
	}
}

/*
** print a line of memory-figures about one process
*/
static void
showmemproc(struct pstat *curstat, double perc)
{
	char	format1[16], format2[16];
	char	format3[16], format4[16];
	char	format5[16], format6[16];
	char	format7[16], format8[16];

	if (curstat->gen.state != 'E')
	{
		/*
		** show process-info of active process
		*/
		printg(memproclina,
		  curstat->gen.pid,
		  val2valstr(curstat->mem.minflt,      format1, 6),
		  val2valstr(curstat->mem.majflt,      format2, 6),
		  val2valstr(curstat->mem.iosw,        format3, 4),
		  val2memstr(curstat->mem.shtext*1024, format4, KBFORMAT),
		  val2memstr(curstat->mem.vmem*1024,   format5, KBFORMAT),
		  val2memstr(curstat->mem.rmem*1024,   format6, KBFORMAT),
		  val2memstr(curstat->mem.vgrow*1024,  format7, KBFORMAT),
		  val2memstr(curstat->mem.rgrow*1024,  format8, KBFORMAT),
		  perc);

		printg(" %.14s\n",   curstat->gen.name);
	}
	else
	{
		/*
		** show process-info of exited process
		*/
		if (curstat->gen.pid == 0)
			printg("    ? ");
		else
			printg("%5d ", curstat->gen.pid);

		printg(memprocline,
		  val2valstr(curstat->mem.minflt,     format1, 6),
		  val2valstr(curstat->mem.majflt,     format2, 6),
		  val2valstr(curstat->mem.iosw,       format3, 4),
		  perc,
		  curstat->gen.name);
	}
}

/*
** print a line of network-figures about one process
*/
static void
shownetproc(struct pstat *curstat, double perc)
{
	char	format1[16], format2[16];
	char	format3[16], format4[16];

	if (curstat->gen.state != 'E')
	{
		/*
		** show process-info of active process
		*/
		printg(netproclina,
		  curstat->gen.pid,
		  val2valstr(curstat->net.tcprcv,      format1, 7),
		  val2valstr(curstat->net.tcpsnd,      format2, 7),
		  val2valstr(curstat->net.udprcv,      format3, 7),
		  val2valstr(curstat->net.udpsnd,      format4, 7),
		  perc);

		printg(" %.14s\n",   curstat->gen.name);
	}
	else
	{
		/*
		** show process-info of exited process
		*/
		if (curstat->gen.pid == 0)
			printg("    ? ");
		else
			printg("%5d ", curstat->gen.pid);

		printg(netprocline,
		  perc,
		  curstat->gen.name);
	}
}

/*
** print a line of (more or less) static info about one process
*/
static void
showstaproc(struct pstat *curstat, double perc)
{
	char		format1[16], format2[16], format3[16];

	struct passwd 	*pwd;
	struct group 	*grp;
	char		*username, *groupname;

	/*
	** determine the user- and group-name of the process
	*/
	if ( (pwd = getpwuid(curstat->gen.uid)) == NULL)
		username = "unknown";
	else
		username = pwd->pw_name;

	if ( (grp = getgrgid(curstat->gen.gid)) == NULL)
		groupname = "unknown";
	else
		groupname = grp->gr_name;

	/*
	** show process-info
	*/
	if (curstat->gen.state != 'E')
	{
		/*
		** show process-info of active process
		*/
		printg(staproclina,
		  	curstat->gen.pid,
		  	username,
		  	groupname,
			convdate(curstat->gen.btime, format1),
			convtime(curstat->gen.btime, format2),
		  	curstat->gen.state,
		  	perc);

		printg(" %.14s\n",   curstat->gen.name);
	}
	else
	{
		/*
		** show process-info of exited process
		*/
		if (curstat->gen.pid == 0)
			printg("    ? ");
		else
			printg("%5d ", curstat->gen.pid);

		if ( curstat->gen.excode&0xff )
			sprintf(format1, "sig_%02d", curstat->gen.excode&0x7f);
		else
		  	val2valstr(curstat->gen.excode>>8, format1, 6);

		printg(staprocline,
		  	username,
		  	groupname,
		  	format1,
			convdate(curstat->gen.btime, format2),
			convtime(curstat->gen.btime, format3),
		  	perc,
		  	curstat->gen.name);
	}
}

/*
** print the system-wide statistics
*/
static int
prisyst(struct sstat *sstat, int curline, int nsecs)
{
	register int	i;
	double		cputot;
	char		format1[16], format2[16], format3[16];
	char		format4[16], format5[16], format6[16];

	/*
	** CPU statistics
	*/
	cputot = (double) sstat->cpu.stime + sstat->cpu.utime +
	                  sstat->cpu.ntime + sstat->cpu.itime;

	if (cputot == 0.0)
		cputot = 1.0;		/* avoid divide-by-zero */

	printg("CPU | system %3.0f%% | user    %3.0f%% | nice %6.0f%% | "
	       "idle %6.0f%% | cs %s |\n",
			(double) (sstat->cpu.stime * 100.0) / cputot,
			(double) (sstat->cpu.utime * 100.0) / cputot,
			(double) (sstat->cpu.ntime * 100.0) / cputot,
			(double) (sstat->cpu.itime * 100.0) / cputot,
			val2valstr(sstat->cpu.csw, format1, 9));

	curline++;

	/*
	** MEMORY statistics
	*/
	printg("MEM | tot  %s | free  %s | cache %s | buff  %s | share %s |\n",
		val2memstr(sstat->mem.physmem  * NBPP, format1, MBFORMAT),
		val2memstr(sstat->mem.freemem  * NBPP, format2, MBFORMAT),
		val2memstr(sstat->mem.cachemem * NBPP, format3, MBFORMAT),
		val2memstr(sstat->mem.buffermem* NBPP, format4, MBFORMAT),
		val2memstr(sstat->mem.sharedmem* NBPP, format5, MBFORMAT));

	curline++;

	/*
	** SWAP statistics
	*/
	printg("SWP | tot  %s | free  %s |              | pin %s | pout %s |\n",
		val2memstr(sstat->mem.totswap  * NBPP, format1, MBFORMAT),
		val2memstr(sstat->mem.freeswap * NBPP, format2, MBFORMAT),
		val2valstr(sstat->mem.pins,            format3, 8),
		val2valstr(sstat->mem.pouts,           format4, 7));

	curline++;

	/*
	** DISK statistics
	**
	** If extended disk-stats's are supported, take them
	*/
	cputot *= 1000.0 / HZ;

	for (i=0; sstat->xdsk[i].name[0]; i++)
	{
		unsigned long 	iotot;

		iotot =  sstat->xdsk[i].nread + sstat->xdsk[i].nwrite;

		if (iotot || fixedhead)
		{
		    printg("DSK | unit %6.6s | busy %6.0lf%% | read %s |"
		           " write %s | avio %4.0lf ms |\n",
		    	sstat->xdsk[i].name,
			sstat->xdsk[i].io_ms * 100.0 / cputot,
		    	val2valstr(sstat->xdsk[i].nread,  format1, 7),
		    	val2valstr(sstat->xdsk[i].nwrite, format2, 6),
		    	iotot ? (double) sstat->xdsk[i].io_ms / iotot : 0.0);

		    curline++;
		}
	}

	/*
	** NET statistics
	*/
	if (sstat->net.tcpin || sstat->net.tcpout ||
	    sstat->net.udpin || sstat->net.udpout || fixedhead )
	{
		printg("NET |   transport | tcpi %s | tcpo %s |"
		       " udpi %s | udpo %s |\n",
			val2valstr(sstat->net.tcpin,   format1, 7),
			val2valstr(sstat->net.tcpout,  format2, 7),
			val2valstr(sstat->net.udpin,   format3, 7),
			val2valstr(sstat->net.udpout,  format4, 7));

		curline++;
	}

	if (sstat->net.ipin || sstat->net.ipout || fixedhead )
	{
		printg("NET |     network | ipi %s | ipo %s | ipfrw %s |"
		       " deliv %s |\n",
			val2valstr(sstat->net.ipin,    format1, 8),
			val2valstr(sstat->net.ipout,   format2, 8),
			val2valstr(sstat->net.ipfrw,   format3, 6),
			val2valstr(sstat->net.ipindel, format4, 6));

		curline++;
	}

	for (i=0; sstat->intf[i].name[0]; i++)
	{
		if (sstat->intf[i].rpack || sstat->intf[i].spack || fixedhead)
		{
			printg("NET | dev %7.7s | pcki %s | pcko %s |"
			       " in %sK | out %sK |\n",
			    	sstat->intf[i].name,
				val2valstr(sstat->intf[i].rpack, format1, 7),
				val2valstr(sstat->intf[i].spack, format2, 7),
				val2valstr(sstat->intf[i].rbyte/1024,
				                                 format3, 8),
				val2valstr(sstat->intf[i].sbyte/1024,
				                                 format4, 7));

			curline++;
		}
	}

	return curline;
}

/*
** generic print-function which checks if printf should be used (to file)
** or curses (to screen)
*/
static int
printg(const char *format, ...)
{
	va_list	args;

	va_start(args, format);

	if (screen)
		vwprintw(stdscr, (char *) format, args);
	else
		vprintf(format, args);

	va_end  (args);
}

/*
** show help information in interactive mode
*/
static char *helptext[] = {
	"Figures shown for active processes:",
	"\t'g' - generic info (default)",
	"\t'm' - memory-details",
	"\t'n' - network-details",
	"\t's' - static process-info",
	"",
	"Sort list of active processes by:",
	"\t'C' - cpu-activity (default)",
	"\t'M' - memory-consumption",
	"\t'D' - disk-activity",
	"\t'N' - network-activity",
	"",
	"Screen-handling:",
	"\t^F  - Forward:  show next     page in the process-list",
	"\t^B  - Backward: show previous page in the process-list",
	"",
	"Miscellaneous commands:",
	"\t'i' - change interval-timer (0 = only manual trigger)",
	"\t't' - manual trigger to finish interval",
	"",
	"\t'u' - focus on specific user",
	"\t'P' - focus on specific process-name",
	"\t'a' - active processes only (default) or all processes (toggle)",
	"\t'p' - pause-button                                     (toggle)",
	"\t'f' - fixate on full range of header-lines             (toggle)",
	"",
	"\t'h' - help-information",
	"\t'?' - help-information",
	"\t'q' - quit this program",
};

static int helplines = sizeof(helptext)/sizeof(char *);

static void
showhelp(int helpline)
{
	int	winlines = LINES-helpline, lin;
	WINDOW	*helpwin;

	/*
	** create a new window for the help-info in which scrolling is
	** allowed
	*/
	helpwin = newwin(winlines, COLS, helpline, 0);
	scrollok(helpwin, 1);

	/*
	** show help-lines 
	*/
	for (lin=0; lin < helplines; lin++)
	{
		wprintw(helpwin, "%s\n", helptext[lin]);

		/*
		** when the window is full, start paging interactively
		*/
		if (lin >= winlines-2)
		{
			wmove    (helpwin, winlines-1, 0);
			wclrtoeol(helpwin);
			wprintw  (helpwin, "Press any key for next line or "
			                   "'q' to leave help .......");

			if (wgetch(helpwin) == 'q')
			{
				delwin(helpwin);
				return;
			}

			wmove  (helpwin, winlines-1, 0);
		}
	}

	wmove    (helpwin, winlines-1, 0);
	wclrtoeol(helpwin);
	wprintw  (helpwin, "End of help - press any key to continue....");
	(void) wgetch(helpwin);
	delwin   (helpwin);
}

/*
** function to be called to print error-messages
*/
void
generic_error(const char *format, ...)
{
	va_list	args;

	va_start(args, format);
	vfprintf(stderr, format, args);
	va_end  (args);
}

/*
** function to be called when the program stops
*/
void
generic_end(void)
{
	endwin();
}

/*
** function to be called when usage-info is required
*/
void
generic_usage(void)
{
	fprintf(stderr, "\t  -g      show general process-info (default)\n");
	fprintf(stderr, "\t  -m      show memory-related process-info\n");
	fprintf(stderr, "\t  -n      show network-related process-info\n");
	fprintf(stderr, "\t  -s      show static process-info\n");
	fprintf(stderr, "\t  -C      sort by cpu-consumption (default)\n");
	fprintf(stderr, "\t  -M      sort by memory-consumption\n");
	fprintf(stderr, "\t  -D      sort by disk-activity\n");
	fprintf(stderr, "\t  -N      sort by network-activity\n\n");
	fprintf(stderr, "\t  -a      show all processes each sample\n");
	fprintf(stderr, "\t  -f      show fixed system-statistic lines\n");
}


/*
** sort-functions
*/
int
compcpu(const void *a, const void *b)
{
	register ulong	acpu = ((struct pstat *)a)->cpu.stime +
			       ((struct pstat *)a)->cpu.utime;
	register ulong	bcpu = ((struct pstat *)b)->cpu.stime +
			       ((struct pstat *)b)->cpu.utime;

	if (acpu < bcpu)
		return(1);
	if (acpu > bcpu)
		return(-1);

	return compmem(a, b);
}

int
compdsk(const void *a, const void *b)
{
	int		i;
	register ulong	adsk = ((struct pstat *)a)->dsk.ior +
			       ((struct pstat *)a)->dsk.iow;
	register ulong	bdsk = ((struct pstat *)b)->dsk.ior +
			       ((struct pstat *)b)->dsk.iow;

	if (adsk < bdsk)
		return(1);
	if (adsk > bdsk)
		return(-1);

	return compcpu(a, b);
}

int
compmem(const void *a, const void *b)
{
	register ulong	amem = ((struct pstat *)a)->mem.rmem;
	register ulong	bmem = ((struct pstat *)b)->mem.rmem;

	if (amem < bmem)
		return(1);
	if (amem > bmem)
		return(-1);
	return 0;
}

int
compnet(const void *a, const void *b)
{
	int		i;
	register ulong	anet = ((struct pstat *)a)->net.tcpsnd +
			       ((struct pstat *)a)->net.tcprcv +
			       ((struct pstat *)a)->net.udpsnd +
			       ((struct pstat *)a)->net.udprcv  ;
	register ulong	bnet = ((struct pstat *)b)->net.tcpsnd +
			       ((struct pstat *)b)->net.tcprcv +
			       ((struct pstat *)b)->net.udpsnd +
			       ((struct pstat *)b)->net.udprcv  ;

	if (anet < bnet)
		return(1);
	if (anet > bnet)
		return(-1);

	return compcpu(a, b);
}
